/*
 * spi.c
 *
 * =============================================================================================
 *  Created on: 21 Aug 2011
 *      Author: Mike McPherson (mmcp42)
 *       email: mmcp42@yahoo.com
 *     version: 1.01
 * =============================================================================================
 *  License - usual free stuff - caveat emptor
 *  if you like it - buy me a beer
 *  if you don't - then I don't know you!
 *
 *  if you do make changes - do let me know so I can keep it up to date
 * =============================================================================================
 */

#include <spi.h>

//==============================================================================
// private function redundant but convenient
//==============================================================================

void spiDoIdle(void);

//==============================================================================

void spiDoIdle(void)
{
	// service stuff we have to do to keep the world spinning
	//=======================================================
	// boardService();
	// usbComService();
}

uint8 spiDeselect(uint8 usartID)
{
	// deselect the nominated SPI device by setting SELECT line HIGH
	// returns PORT value
	// returns 0 on error
	//==============================================================

	uint8 alternate;
	uint8 port;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiDeselect USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	if (usartID == 0)
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 1;

		// de-select the device
		//=====================
		if (alternate == 0)
		{
			// port 0 bit 4
			//=============
			port = P0;
			debug("spiDeselect USART0 alternate 1 P0: %x\n", port);

			P0 |= 0x10;
			port = P0;
			debug("spiDeselect USART0 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 2
			//=============
			port = P1;
			debug("spiDeselect USART0 alternate 2 P1: %x\n", port);

			P1 |= 0x04;
			port = P1;
			debug("spiDeselect USART0 alternate 2 P1: %x\n", port);
		}
	}
	else
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 2;

		// de-select the device
		//=====================
		if (alternate == 0)
		{
			// port 0 bit 2
			//=============
			port = P0;
			debug("spiDeselect USART1 alternate 1 P0: %x\n", port);

			P0 |= 0x04;
			port = P0;
			debug("spiDeselect USART1 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 4
			//=============
			port = P1;
			debug("spiDeselect USART1 alternate 2 P1: %x\n", port);

			P1 |= 0x10;
			port = P1;
			debug("spiDeselect USART1 alternate 2 P1: %x\n", port);
		}
	}
	return port;
}

uint8 spiInit(uint8 usartID, uint8 alternate, uint8 polarity, uint8 phase, uint8 bitOrder)
{
	// initialise SPI interface
	// return 1 for success
	// return 0 on error
	//=========================

	// USART_ID is 0 or 1
	// alternate is 1 or 2 for the two choices of alternative location
	// polarity is 0 for Negative, 1 for Positive
	// phase is 0 for Negative 1 for Positive
	// bitOrder is 0 for MSB first, 1 for LSB first
	// defines for easier reading:
	//
	//	spiUsart0 0
	//	spiUsart1 1
	//
	//	spiAlternate1 1
	//	spiAlternate2 2
	//
	//	spiPhaseN 0
	//	spiPhaseP 1
	//
	//	spiPolarityN 0
	//	spiPolarityP 1
	//
	//	spiBitOrderM 0
	//	spiBitOrderL 1
	//================================================================

	// range check usartID
	//====================
	if (usartID > spiUsart1)
	{
		debug("spiInit USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// range check alternate
	//=====================
	if (alternate < spiAlternate1 || alternate > spiAlternate2)
	{
		debug("alternate %d out of range[1-2]\n", alternate);
		return 0;
	}

	// range check polarity
	//=====================
	if (polarity > spiPolarityP)
	{
		debug("polarity %d out of range[0-1]\n", polarity);
		return 0;
	}

	// range check phase
	//==================
	if (phase > spiPhaseP)
	{
		debug("phase %d out of range[0-1]\n", phase);
		return 0;
	}

	// range check bitOrder
	//=====================
	if (bitOrder > spiBitOrderL)
	{
		debug("bitOrder %d out of range[0-1]\n", bitOrder);
		return 0;
	}

	if (usartID == spiUsart0)
	{
		// process USART0
		//===============
		debug("USART 0\n");

		// start by clearing U0CSR
		//========================
		U0CSR = 0;

		// set SPI mode
		//=============
		U0CSR &= ~0x80;	// U0CSR bit 7=0 is SPI mode
		debug("spiInit SPI mode U0CSR.MODE: %2x\n", U0CSR & 0x80);

		// set Master Mode
		//================
		U0CSR &= ~0x20;	// U0CSR bit 5=0 is Master mode
		debug("spiInit Master mode U0CSR: %2x\n", U0CSR & 0x20);

		if (alternate == spiAlternate1)
		{
			// set USART0 alternate 1 pins
			//============================
			PERCFG &= ~0x01;	// USART0 alternate 1
			P0SEL |= 0x2C;  	// P0_5, P0_3, and P0_2 are peripherals
			P0SEL &= ~0x10; 	// P0_4 is GPIO (SSN)
			P0DIR |= 0x10;  	// SSN is set as output

			debug("spiInit alternate 1 PERCFG.U0CFG: %2x\n", PERCFG & 0x01);
			debug("spiInit alternate 1 P0SEL: %2x\n", P0SEL);
			debug("spiInit alternate 1 P0DIR: %2x\n", P0DIR);
		}
		else
		{
			// set USART0 alternate 2 pins
			//============================
			PERCFG |= 0x01;		// USART0 alternate 2
			P1SEL |= 0x38;  	// P1_3, P1_4, and P1_5 are peripherals
			P1SEL &= ~0x04; 	// P1_2 is GPIO (SSN)
			P1DIR |= 0x04;  	// SSN is set as output

			debug("spiInit alternate 2 PERCFG.u0CFG: %2x\n", PERCFG & 0x01);
			debug("spiInit alternate 2 P1SEL: %2x\n", P1SEL);
			debug("spiInit alternate 2 P1DIR: %2x\n", P1DIR);
		}

		// Set baud rate to max (system clock frequency / 8)
		//==================================================
		spiSetSpeed(usartID, 3000000);
		debug("spiInit U0BAUD_M: %2x\n", U0BAUD);

		// SPI polarity
		//=============
		if (polarity == spiPolarityP)
			U0GCR |= 0x80;
		else
			U0GCR &= ~0x80;

		// SPI phase
		//==========
		if (phase == spiPhaseP)
			U0GCR |= 0x40;
		else
			U0GCR &= ~0x40;

		// SPI bit order
		//==============
		if (bitOrder == spiBitOrderM)
			U0GCR |= 0x20;
		else
			U0GCR &= ~0x20;

		debug("spiInit U0GCR.CPOLR:  %2x\n", U0GCR & 0x80);
		debug("spiInit U0GCR.CPHA:   %2x\n", U0GCR & 0x40);
		debug("spiInit U0GCR.ORDER:  %2x\n", U0GCR & 0x20);
		debug("spiInit U0GCR.BAUD_E: %2x\n", U0GCR & 0x1F);

		// priority USART 0
		//=================
		P2DIR &= ~0xC0;
		debug("spiInit P2DIR.PRIP0: %2x\n", P2DIR & 0xC0);
	}
	else
	{
		// process USART1
		//===============

		debug("USART1\n");

		// start by clearing U1CSR
		//========================
		U1CSR = 0;

		// set SPI mode
		//=============
		U1CSR &= ~0x80;	// U1CSR bit 7=0 is SPI mode

		debug("spiInit SPI mode U1CSR: %2x\n", U1CSR & 0x80);

		// set Master Mode
		//================
		U1CSR &= ~0x20;	// U1CSR bit 5=0 is Master mode

		debug("spiInit Master mode U1CSR: %2x\n", U1CSR & 0x20);

		if (alternate == spiAlternate1)
		{
			// set USART1 alternate 1 pins
			//============================
			PERCFG &= ~0x02;	// USART1 alternate 1
			P0SEL |= 0x38;  	// P0_5, P0_4, and P0_3 are peripherals
			P0SEL &= ~0x04; 	// P0_2 is GPIO (SSN)
			P0DIR |= 0x04;  	// SSN is set as output

			debug("spiInit alternate 1 PERCFG: %2x\n", PERCFG & 0x02);
			debug("spiInit alternate 1 P0SEL: %2x\n", P0SEL);
			debug("spiInit alternate 1 P0DIR: %2x\n", P0DIR);
		}
		else
		{
			// set USART1 alternate 2 pins
			//============================
			PERCFG |= 0x02;		// USART1 alternate 2
			P1SEL |= 0xE0;  	// P1_7, P1_6, and P1_5 are peripherals
			P1SEL &= ~0x10; 	// P1_4 is GPIO (SSN)
			P1DIR |= 0x10;  	// SSN is set as output

			debug("spiInit alternate 2 PERCFG: %2x\n", PERCFG & 0x02);
			debug("spiInit alternate 2 P1SEL: %2x\n", P1SEL);
			debug("spiInit alternate 2 P1DIR: %2x\n", P1DIR);
		}

		// Set baud rate to max (system clock frequency / 8)
		//==================================================
		spiSetSpeed(usartID, 0);
		debug("spiInit U1BAUD_M: %2x\n", U1BAUD);

		// SPI polarity
		//=============
		if (polarity == spiPolarityP)
			U1GCR |= 0x80;
		else
			U1GCR &= ~0x80;

		// SPI phase
		//==========
		if (phase == spiPhaseP)
			U1GCR |= 0x40;
		else
			U1GCR &= ~0x40;

		// SPI bit order
		//==============
		if (bitOrder == spiBitOrderM)
			U1GCR |= 0x20;
		else
			U1GCR &= ~0x20;

		debug("spiInit U1GCR.CPOLR:  %2x\n", U1GCR & 0x80);
		debug("spiInit U1GCR.CPHA:   %2x\n", U1GCR & 0x40);
		debug("spiInit U1GCR.ORDER:  %2x\n", U1GCR & 0x20);
		debug("spiInit U1GCR.BAUD_E: %2x\n", U1GCR & 0x1F);

		//1st priority: USART 1
		//=====================
		P2SEL |= 0x40;
		debug("spiInit P2DIR.PRIP0: %2x\n", P2DIR & 0xC0);
	}

	return 1;
}

uint8 spiReadByte(uint8 usartID)
{
	// read one byte of data from SPI
	// return byte read
	// return 0 on error
	//===============================

	uint8 buffer[8];
	uint16 size;

	// it's like spiReadData, but only one byte into private buffer
	//=============================================================

	// send command over SPI interface
	// it's like send data, but without the buffer
	// so we fake it!
	//================================

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiReadByte USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// read the data into our private buffer
	// then return it
	//======================================
	buffer[0] = 0xA5;
	size = spiReadData(usartID, buffer, 1);
	if (size == 1)
		return buffer[0];
	else
		return 0;
}

uint16 spiReadData(uint8 usartID, uint8* buffer, uint16 size)
{
	// read data from SPI
	// return amount of data read
	// return errcode on error
	// errcode 0x8001 = usartID out of range
	// errcode 0x8002 = data timeout
	//====================================

	uint16 i;
	uint32 startTime;
	uint32 timeSpent;
	uint32 idleCount;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiReadData USART %d out of range[0-1]\n", usartID);
		return 8001;
	}

	// note current (ms) time
	//=======================
	startTime = getMs();
	idleCount = 0;

	// read data from SPI into buffer
	//===============================
	if (usartID == spiUsart0)
	{
		for (i = 0; i < size; i++)
		{
			// clear receive flags
			//====================
			U0CSR &= ~0x06;

			// write empty data to the device
			//===============================
			U0DBUF = 0xFF;

			// wait for data transmit flag
			//===========================
			while (!(U0CSR & 0x02))
			{
				idleCount++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiReadData timeout\n");
					debug("spiReadData idleCount %ld\n", idleCount);
					return 0x8002;
				}
			}
			// read data into the buffer
			//==========================
			buffer[i] = U0DBUF;
		}
	}
	else
	{
		for (i = 0; i < size; i++)
		{
			// clear receive flag
			//===================
			U1CSR &= ~0x06;

			// write empty data to the device
			//===============================
			U1DBUF = 0xFF;
			idleCount = 0;

			// wait for data transmit flag
			//===========================
			while (!(U1CSR & 0x02))
			{
				idleCount++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiReadData timeout\n");
					debug("spiReadData idleCount %ld\n", idleCount);
					return 0x8002;
				}
			}

			// read data into the buffer
			//==========================
			buffer[i] = U1DBUF;
		}
	}

	debug("spiReadData idleCount %ld\n", idleCount);
	return size;
}

uint8 spiSelect(uint8 usartID)
{
	// select the nominated SPI device by setting SELECT line HIGH
	// return PORT contents if ok
	// return 0 on error
	//============================================================

	uint8 alternate;
	uint8 port;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiSelect USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	if (usartID == 0)
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 1;

		// select the device
		//==================
		if (alternate == 0)
		{
			// port 0 bit 4
			//=============
			port = P0;
			debug("spiSelect USART0 alternate 1 P0: %x\n", port);

			P0 &= ~0x10;
			port = P0;
			debug("spiSelect USART0 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 2
			//=============
			port = P1;
			debug("spiSelect USART0 alternate 2 P1: %x\n", port);

			P1 &= ~0x04;
			port = P1;
			debug("spiSelect USART0 alternate 2 P1: %x\n", port);
		}
	}
	else
	{
		// get the alternate so we know which port for SS
		//===============================================
		alternate = PERCFG & 2;

		// select the device
		//==================
		if (alternate == 0)
		{
			// port 0 bit 2
			//=============
			port = P0;
			debug("spiSelect USART1 alternate 1 P0: %x\n", port);

			P0 &= ~0x04;
			port = P0;
			debug("spiSelect USART1 alternate 1 P0: %x\n", port);
		}
		else
		{
			// port 1 bit 4
			//=============
			port = P1;
			debug("spiSelect USART1 alternate 2 P1: %x\n", port);

			P1 &= ~0x10;
			port = P1;
			debug("spiSelect USART1 alternate 2 P1: %x\n", port);
		}
	}
	return port;
}

uint32 spiSetSpeed(uint8 usartID, uint32 speed)
{
    uint32 baudMPlus256;
    uint8 baudE = 0;
    uint32 freq;

	// set SPI speed
	// return speed set
	// return 0 on error
	//==================
    freq = speed;

    // max baud rate is 3000000 (F/8)
    // min is 23 (baudM = 1)
    //===============================
    if (freq < 23 || freq > 3000000)
        return 0;

    // 495782 is the largest value that will not overflow the following calculation
    //=============================================================================
    while (freq > 495782)
    {
        baudE++;
        freq /= 2;
    }

    // calculate baud rate - see datasheet 12.14.3
    // this is derived from (baudM + 256) = baud * 2^28 / 24000000
    //============================================================
    baudMPlus256 = (freq * 11) + (freq * 8663 / 46875);

    // get baudMPlus256 into the range 256-511 (so BAUD_M is in the range 0-255)
    // UNBAUD.BAUD_M (7:0) - only the lowest 8 bits of baudMPlus256 are used,
    // so this is effectively baudMPlus256 - 256
    //==========================================================================
    while (baudMPlus256 > 0x1ff)
    {
        baudE++;
        baudMPlus256 /= 2;
    }

    debug("spiSetSpeed BAUDE: %d, BUADM: %ld\n", baudE, baudMPlus256 & 0xFF);
    if (usartID == 0)
	{
		// set USART0 speed
		//=================
		U0BAUD = baudMPlus256 & 0xFF;
		U0GCR &= 0xE0;
		U0GCR |= (baudE & 0x1F);
	}
	else
	{
		// set USART1 speed
		//=================
		U1BAUD = baudMPlus256 & 0xFF;
		U1GCR &= 0xE0;
		U1GCR |= (baudE & 0x1F);
	}
	return speed;
}

uint16 spiWriteByte(uint8 usartID, uint8 byte)
{
	// write a single byte
	// return 1 for success
	// return 0 on error
	//=====================

	uint8 buffer[8];
	uint16 size;

	// it's like spiWriteData, but only one byte into private buffer
	//==============================================================

	// send command over SPI interface
	// it's like write data, but without the buffer
	// so we fake it!
	//=============================================

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiWriteByte USART %d out of range[0-1]\n", usartID);
		return 0;
	}

	// write the data from our private buffer
	//======================================
	buffer[0] = byte;
	size = spiWriteData(usartID, buffer, 1);
	return size;
}

uint16 spiWriteData(uint8 usartID, uint8* buffer, uint16 size)
{
	// write buffer full of data to SPI device
	// return amount of data written
	// return errcode on error
	// errcode 0x8001 = usartID out of range
	// errcode 8002 = data timeout
	//========================================
	uint16 i;
	uint32 startTime;
	uint32 timeSpent;
	uint32 tries;

	// range check USART_ID
	//=====================
	if (usartID > spiUsart1)
	{
		debug("spiWriteData USART %d out of range[0-1]\n", usartID);
		return 0x8001;
	}

	// dump buffer contents
	//=====================
	debug("spiWriteData USART%d %d bytes\n", usartID, size);

	// note current (ms) time
	//=======================
	startTime = getMs();
	tries = 0;

	// write data from buffer to SPI
	//==============================
	if (usartID == spiUsart0)
	{
		// loop writing data
		//==================
		for (i = 0; i < size; i++)
		{
			// clear data transmitted flag
			//============================
			U0CSR &= ~0x02;

			// send a byte
			//============
			U0DBUF = buffer[i];

			// wait for data transmitted
			//==========================
			while (!(U0CSR & 0x02))
			{
				tries++;
				spiDoIdle();
				{
					timeSpent = getMs() - startTime;
					if (timeSpent > TIMEOUT)
					{
						// took too long to respond
						//=========================
						debug("spiWriteData timeout\n");
						return 0x8002;
					}
				}
			}
		}
	}
	else
	{
		// loop writing data
		//==================
		for (i = 0; i < size; i++)
		{
			// clear data transmitted flag
			//============================
			U1CSR &= ~0x02;

			// send a byte
			//============
			U1DBUF = buffer[i];

			// wait for data transmitted
			//==========================
			while (!(U1CSR & 0x02))
			{
				tries++;
				spiDoIdle();
				timeSpent = getMs() - startTime;
				if (timeSpent > TIMEOUT)
				{
					// took too long to respond
					//=========================
					debug("spiWriteData timeout\n");
					return 0x8002;
				}
			}
		}
	}

	debug("spWriteData tries: %ld\n", tries);

	return size;
}
